# Copyright 2023 Luminary Cloud, Inc. All Rights Reserved.
from datetime import datetime
from typing import BinaryIO, List, Optional, cast
import io
import tarfile

from ._wrapper import ProtoWrapper, ProtoWrapperBase
from ._client import get_default_client
from ._helpers.timestamp_to_datetime import timestamp_to_datetime
from ._helpers.wait_for_simulation import wait_for_simulation
from ._helpers import download_surface_solution, download_volume_solution
from ._proto.simulation import simulation_pb2 as simulationpb
from ._proto.common import common_pb2 as commonpb
from .enum import (
    AveragingType,
    CalculationType,
    QuantityType,
    ResidualNormalization,
    SimulationStatus,
)
from .vector3 import Vector3


class _DownloadedTextFile(io.StringIO):
    filename: str

    def __init__(self, file_proto: commonpb.File):
        super().__init__(file_proto.data.decode())
        metadata = file_proto.metadata
        self.filename = metadata.name
        if metadata.ext:
            self.filename += "." + metadata.ext


@ProtoWrapper(simulationpb.Simulation)
class Simulation(ProtoWrapperBase):
    """Represents a simulation object."""

    id: str
    "Simulation ID."
    name: str
    "Simulation name."
    status: SimulationStatus
    "Simulation status. May not reflect current status."
    mesh_id: str
    "ID of the simulation mesh."

    _proto: simulationpb.Simulation

    @property
    def create_time(self) -> datetime:
        return timestamp_to_datetime(self._proto.create_time)

    @property
    def update_time(self) -> datetime:
        return timestamp_to_datetime(self._proto.update_time)

    def update(
        self,
        *,
        name: str = None,
    ) -> None:
        """
        Update/Edit simulation attributes.

        Mutates self.

        Parameters
        ----------
        name : str
            (Optional) New project name.
        """
        req = simulationpb.UpdateSimulationRequest(
            id=self.id,
            name=name,
        )
        res = get_default_client().UpdateSimulation(req)
        self._proto = res.simulation

    def wait(
        self,
        *,
        print_residuals: bool = False,
        interval_seconds: float = 5,
        timeout_seconds: float = float("inf"),
    ) -> SimulationStatus:
        """
        Wait until the simulation is completed, failed, or suspended.

        Parameters
        ----------
        print_residuals: bool
            (Optional) If true, residual values for the latest completed iteration will be printed.
            Frequency is based on interval_seconds.
        interval : float
            (Optional) Number of seconds between polls.
        timeout : float
            (Optional) Number of seconds before timeout.

        Returns
        -------
        luminarycloud.enum.SimulationStatus
            Current status of the simulation.
        """
        wait_for_simulation(
            get_default_client(),
            self._proto,
            print_residuals=print_residuals,
            interval_seconds=interval_seconds,
            timeout_seconds=timeout_seconds,
        )
        self._proto = get_simulation(self.id)._proto
        return self.status

    def download_global_residuals(
        self,
        normalization: ResidualNormalization = ResidualNormalization.RELATIVE,
        all_iterations: bool = False,
    ) -> _DownloadedTextFile:
        """
        Download global residuals in csv format.

        Parameters
        ----------
        normalization : ResidualNormalization
            (Optional) The type of normalization to use. Default is relative normalization.
        all_iterations : bool
            (Optional) If true, all iteration values are output. By default, only the final
            iteration value is output. This can only be set for steady state simulations.

        Returns
        -------
        io.StringIO
            Stream of text. The filename can be retrieved from the "filename" attribute of the
            object.

        Examples
        --------
        Create a Pandas dataframe:

        >>> from luminarycloud.enum import ResidualType
        >>> import pandas as pd
        >>> with simulation.download_global_residuals() as dl:
        ...     residuals_df = pd.read_csv(dl)

        Save to disk:

        >>> with download_global_residuals() as dl:
        ...     with open(dl.filename, "wb") as fp:
        ...         fp.write(dl.read())
        """
        req = simulationpb.GetSimulationGlobalResidualsRequest(
            id=self.id,
            residual_normalization=normalization.value,
            all_iterations=all_iterations,
        )
        res = get_default_client().GetSimulationGlobalResiduals(req)
        return _DownloadedTextFile(res.csv_file)

    def download_surface_output(
        self,
        quantity_type: QuantityType,
        surface_ids: List[str],
        calculation_type: CalculationType = CalculationType.PER_SURFACE,
        frame_id: str = "",
        force_direction: Optional[Vector3] = None,
        moment_center: Optional[Vector3] = None,
        averaging_type: AveragingType = AveragingType.UNSPECIFIED,
        all_iterations: bool = False,
    ) -> _DownloadedTextFile:
        """
        Downloads surface outputs (e.g. lift, drag, ...) in csv format.

        Parameters
        ----------
        quantity_type : luminarycloud.enum.QuantityType
            Surface quantity type to compute (e.g. lift, drag).
        surface_ids : list of str
            List of names of the surfaces to compute the quantities for.
            Should have at least one element.
        calculation_type : CalculationType
            Whether the calculation should be done for all the surfaces together or each surface
            individually. Default is PER_SURFACE.
        frame_id: str
            (Optional) The ID of the reference frame that this output should be
            reported in for "force" quantity types.
        force_direction : Vector3
            (Optional) The direction of the query component for "force" quantity types.
            Required for certain quantity types.
        moment_center : Vector3
            (Optional) The center of moment for "force" quantity types.
            Required for certain quantity types. Ignored if not applicable.
        averaging_type : AveragingType
            (Optional) The averaging method used to compute "surface average" quantity types.
            Ignored if not applicable.
        all_iterations : bool
            (Optional) If true, all iteration values are output. By default, only the final
            iteration value is output. This can only be set for steady state simulations.

        Returns
        -------
        io.StringIO
            Stream of text. The filename can be retrieved from the "filename" attribute of the
            object.

        Examples
        --------
        Create a Pandas dataframe:

        >>> from luminarycloud.enum import QuantityType
        >>> import pandas as pd
        >>> with simulation.download_surface_quantity(QuantityType.LIFT, ["0/bound/airfoil"], frame_id="body_frame_id") as dl:
        ...     residuals_df = pd.read_csv(dl)

        Save to disk:

        >>> with simulation.download_surface_quantity(QuantityType.LIFT, ["0/bound/airfoil"]) as dl:
        ...     with open(dl.filename, "wb") as fp:
        ...         fp.write(dl.read())
        """
        req = simulationpb.GetSimulationSurfaceQuantityOutputRequest(
            id=self.id,
            quantity_type=quantity_type.value,
            surface_ids=surface_ids,
            calculation_type=calculation_type.value,
            frame_id=frame_id,
            force_direction=force_direction._to_proto() if force_direction else None,
            moment_center=moment_center._to_proto() if moment_center else None,
            averaging_type=averaging_type.value,
            all_iterations=all_iterations,
        )
        res = get_default_client().GetSimulationSurfaceQuantityOutput(req)
        return _DownloadedTextFile(res.csv_file)

    def download_surface_solution(self, time_step: Optional[int] = None) -> tarfile.TarFile:
        """
        Download the raw surface data as a gzipped tarball containing .vtu files.

        Parameters
        ----------
        time_step : int
            The time step of the solution to download. Required for transient simulations.
            Should be None otherwise.

        Returns
        -------
        tarfile.Tarfile

        Examples
        --------
        >>> with simulation.download_surface_solution() as streaming_tar_file:
        ...     path = f"./surface_data_{simulation.id}"
        ...     streaming_tar_file.extractall(path)
        ...     print(f"Extracted files to {path}:")
        ...     print("\\t" + "\\n\\t".join(os.listdir(path)))
        Extracted files to ./surface_data_sim-ac104e33-2784-4993-a291-ebf9e6cd3b97:
            surface_0_bound_z_minus.vtu
            summary.vtm
            surface_0_bound_airfoil.vtu
            surface_0_bound_z_plus.vtu
            surface_0_bound_farfield.vtu
        """
        stream = download_surface_solution(
            get_default_client(),
            simulation_id=self.id,
            time_step=time_step,
        )
        return tarfile.open(
            name=stream.filename,
            fileobj=cast(BinaryIO, stream),
            mode="r|gz",
        )

    def delete(self) -> None:
        """
        Delete the simulation.

        The simulation will be stopped first if running. This operation cannot be reverted and all
        the simulation data will be deleted as part of this request.
        """
        req = simulationpb.DeleteSimulationRequest(id=self.id)
        get_default_client().DeleteSimulation(req)

    def suspend(self) -> None:
        """
        Suspend the simulation.
        """
        req = simulationpb.SuspendSimulationRequest(id=self.id)
        get_default_client().SuspendSimulation(req)

    def download_volume_solution(self) -> tarfile.TarFile:
        """
        Download volume solution for a completed steady simulation as a gzipped tarball containing .vtu & .vtp files.

        The output may be broken up into multiple .vtu files for large simulations.

        Returns
        -------
        tarfile.Tarfile

        Examples
        --------
        >>> with simulation.download_volume_solution() as streaming_tar_file:
        ...     path = f"./volume_solution_{simulation.id}"
        ...     streaming_tar_file.extractall(path)
        ...     print(f"Extracted files to {path}:")
        ...     for root, dirs, filenames in os.walk(path):
        ...         print("\\t" + "\\n\\t".join([os.path.join(root, file) for file in filenames]))
        Extracted files to ./volume_solution_<simulation.id>:
                ./volume_solution_<simulation.id>/volume_solution_<simulation.id>.vtm
                ./volume_solution_<simulation.id>/volume_solution_<simulation.id>/volume_solution_<simulation.id>_1_0.vtp
                ./volume_solution_<simulation.id>/volume_solution_<simulation.id>/volume_solution_<simulation.id>_3_0.vtp
                ./volume_solution_<simulation.id>/volume_solution_<simulation.id>/volume_solution_<simulation.id>_4_0.vtp
                ./volume_solution_<simulation.id>/volume_solution_<simulation.id>/volume_solution_<simulation.id>_2_0.vtp
                ./volume_solution_<simulation.id>/volume_solution_<simulation.id>/volume_solution_<simulation.id>_0_0.vtu
        """
        stream = download_volume_solution(
            get_default_client(),
            simulation_id=self.id,
        )
        return tarfile.open(
            name=stream.filename,
            fileobj=cast(BinaryIO, stream),
            mode="r|gz",
        )


def get_simulation(id: str) -> Simulation:
    """
    Retrieve a specific simulation by ID.

    Parameters
    ----------
    id : str
        Simulation ID.
    """
    req = simulationpb.GetSimulationRequest(id=id)
    res = get_default_client().GetSimulation(req)
    return Simulation(res.simulation)
